/* ddh.h - functions to read ECMWF DDH *var* files */

#include <stdio.h>
#include "ddh.h"

#define MAX_LINE_LENGTH 256

/* Fatal errors */
static T_ERROR ERR_NOHASH = {1, "First character is not \"#\"", TRUE};
static T_ERROR ERR_TRUNCATED = {2, "Truncated file", TRUE};
static T_ERROR ERR_CORRUPT = {3, "File corrupted", TRUE};
static T_ERROR ERR_TOOMANYLEVELS = {4, "Too many levels", TRUE};
static T_ERROR ERR_TOOMANYPROFILES = {5, "Too many profiles", TRUE};

/* Non-fatal errors */
static T_ERROR ERR_INITCHANGE = {6, "Data taken from more than one forecast", FALSE};

T_ERROR *
ddh_read(FILE *file, DDHday *data)
{
  float itim, vtim;
  float Rgas = 287.0; /* Gas constant */
  float g = 9.81;  /* Acceleration due to gravity */
  float epsilon = 0.622;
  float epsilon_factor = 1.0 - 1.0/epsilon;
  int ch;
  int profile = -1, lev = 0, ilev = 0;
  int idat, vdat;
  char buffer[MAX_LINE_LENGTH];
  char warn_init_change = 0;

  ch = fgetc(file);
  if (ch != '#') {
    return &ERR_NOHASH;
  }

  if (!fgets(buffer, MAX_LINE_LENGTH, file)) {
    return &ERR_TRUNCATED;
  }

  if (!fgets(buffer, MAX_LINE_LENGTH, file)) {
    return &ERR_TRUNCATED;
  }
  if (sscanf(buffer+9, "%d %*d %f %f",
	     &data->station_id, &data->longitude, &data->latitude) < 3) {
    return &ERR_CORRUPT;
  }
  if (data->station_id == 21) {
    data->station_name = "Chilbolton";
  }
  else if (data->station_id == 30) {
    data->station_name = "Cabauw";
  }
  else if (data->station_id == 47) {
    data->station_name = "Palaiseau";
  }
  else {
    data->station_name = NULL;
  }

  if (!fgets(buffer, MAX_LINE_LENGTH, file)) {
    return &ERR_TRUNCATED;
  }
  data->nprofiles = 0;
  data->nlevels = 0;
  data->idat = 0;
  data->itim = 0;
  data->sigma0 = -1;

  while ((ch = fgetc(file)) != '#') {
    float P[DDH_MAX_LEVELS];
    float U[DDH_MAX_LEVELS];
    float V[DDH_MAX_LEVELS];
    float T[DDH_MAX_LEVELS];
    float Q[DDH_MAX_LEVELS];
    float L[DDH_MAX_LEVELS];
    float I[DDH_MAX_LEVELS];
    float A[DDH_MAX_LEVELS];
    float R[DDH_MAX_LEVELS];
    float W[DDH_MAX_LEVELS];
    float H[DDH_MAX_LEVELS];
    if (ch < 0) {
      break;
    }

    ungetc(ch, file);

    if (!fgets(buffer, MAX_LINE_LENGTH, file)) {
      return &ERR_TRUNCATED;
    }

    if (sscanf(buffer, "%d %f %d %f %d", &idat, &itim, &vdat, &vtim, &lev)
	< 5) {
      return &ERR_CORRUPT;
    }

    if (lev > DDH_MAX_LEVELS) {
      return &ERR_TOOMANYLEVELS;
    }

    if (!data->idat) {
      data->idat = idat;
      data->itim = itim;
    }
    else if (data->idat != idat || data->itim != itim) {
      warn_init_change = 1;
    }

    if (profile < 0) {
      data->vdat[0] = vdat;
      data->vtim[0] = vtim;
      profile = 0;
    }
    else if (data->vdat[profile] != vdat || data->vtim[profile] != vtim) {
      ++profile;
      if (profile >= DDH_MAX_PROFILES) {
	return &ERR_TOOMANYPROFILES;
      }
      data->vdat[profile] = vdat;
      data->vtim[profile] = vtim;
    }
    ilev = lev-1;
    if (sscanf(buffer+34, "%e %e %e %e %e %e %e %e %e %e",
	       &P[ilev], &U[ilev], &V[ilev], &T[ilev], &Q[ilev],
	       &L[ilev], &I[ilev], &A[ilev], &R[ilev], &W[ilev])
	< 10) {
      return &ERR_CORRUPT;
    }

    if (data->sigma0 < 0.0) {
      /*      data->sigma0 = 0.99164;*/
      data->sigma0 = 0.998812;
    }

    /* Calculate heights */
    H[ilev] = (1.0/data->sigma0 - 1.0) * Rgas * T[ilev]
      / (g * (1.0 + epsilon_factor * Q[ilev]));
    for (ilev = ilev-1 ; ilev >= 0; ilev--) {
      H[ilev] = H[ilev+1]
	+ (P[ilev+1] - P[ilev]) * Rgas
	* (T[ilev] + T[ilev+1])
	/ (g * (P[ilev] + P[ilev+1])
	   * (1.0 + epsilon_factor * (Q[ilev+1] + Q[ilev]) * 0.5));
    }

    for (ilev = 0; ilev < lev; ilev++) {
      int level = lev - ilev - 1;
      data->P[profile][ilev] = P[level];
      data->U[profile][ilev] = U[level];
      data->V[profile][ilev] = V[level];
      data->T[profile][ilev] = T[level];
      data->Q[profile][ilev] = Q[level];
      data->L[profile][ilev] = L[level];
      data->I[profile][ilev] = I[level];
      data->A[profile][ilev] = A[level];
      data->R[profile][ilev] = R[level];
      data->W[profile][ilev] = W[level];
      data->H[profile][ilev] = H[level];
    }
  }
  ungetc('#', file);
  data->nprofiles = profile + 1;
  data->nlevels = lev;

  if (warn_init_change) {
    return &ERR_INITCHANGE;
  }

  return NULL;
}
