#include <stdlib.h>
#include <string.h>
#include <netcdf.h>

#include "ddh_attenuation.h"
#include "nctools.h"
#include "version.h"

#define DEBUG fprintf(stderr, "Line: %d of %s\n",  __LINE__, __FILE__)
#define MAX_FREQS 10
#define MAX_STRING_LENGTH 512
static int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); nc_close(ncid); exit(-1); \
} 
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
void
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
void
add_attributes(int ncid, int fieldid, char *units,
	       char *long_name, char *standard_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
  if (standard_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "standard_name",
				 strlen(standard_name), standard_name));
  }
}

static
void
add_float_array(int ncid, int *fieldid, char *name, int *dimids,
		char *units, char *long_name, char *standard_name)
{
  report_error(nc_def_var(ncid, name, NC_FLOAT, 2, dimids, fieldid),
	       name);
  add_attributes(ncid, *fieldid, units, long_name, standard_name);
}

static
void
add_3D_float_array(int ncid, int *fieldid, char *name, int *dimids,
		char *units, char *long_name, char *standard_name)
{
  report_error(nc_def_var(ncid, name, NC_FLOAT, 3, dimids, fieldid),
	       name);
  add_attributes(ncid, *fieldid, units, long_name, standard_name);
}

static
void
usage(char **argv)
{
  fprintf(stderr,
	  "Usage: %s [options] ddhfile1 [ddhfile2 ...]\n"
	  "Options:\n"
	  "  -f[req] f    calculate radar/radiometer parameters at f GHz\n"
	  "                 (this argument may be called up to %d times)\n"
	  "    -user s    specify the user and their email address\n",
	  argv[0], MAX_FREQS);   
}

int
main(int argc, char **argv)
{
  int iarg = 1, ifile;
  char *user = NULL;
  float freq[MAX_FREQS];
  int nfreqs = 0;
  char quiet = 0;

  /* Parse switches */
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  fprintf(stderr, 
		  "Warning: user specified more than once: `%s' not used\n",
		  user);
	}
        user = argv[++iarg];
      }
      else {
        fprintf(stderr, "Error: user not specified\n");
        exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-f") == 0
	     || strcmp(argv[iarg], "-freq") == 0 ) {
      if (argc > iarg+1) {
	if (nfreqs >= MAX_FREQS) {
	  fprintf(stderr, "Too many frequencies: ignoring %s GHz\n", argv[++iarg]);
	}
	freq[nfreqs++] = atof(argv[++iarg]);
	if (freq[nfreqs-1] <= 0.0) {
	  fprintf(stderr, "Error: frequency must be positive and non zero\n");
	  exit(1);
	}
      }
      else {
        fprintf(stderr, "Error: frequency not specified\n");
        exit(1);
      }
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    usage(argv);
    exit(0);
  }

  for (ifile = iarg; ifile < argc; ++ifile) {
    size_t tmp_index;
    FILE *file;
    DDHday data;
    T_ERROR *err;
    char tmp_buffer[MAX_STRING_LENGTH];
    int i, j;
    int ncid;
    int timedimid, leveldimid, freqdimid;
    int timeid, levelid, freqid;
    int latid, lonid;
    int varid[11];
    int dimids[3];
    int attenid, attenrateid;
    int satattenrateid;
    int K2id, liquidattenrateid;
    int year, month, day;
    int hour, minute, second;
    float hour_mod = 0.0; /* usually 0 or 24 */

    file = fopen(argv[ifile], "r");
    if (!file) {
      fprintf(stderr, "Error opening %s\n", argv[ifile]);
      continue;
    }

    err = ddh_read(file, &data);
    T_ERROR_CHECK(err);

    /* We have read a valid ECMWF data file */
    /* Create a netcdf file */
    snprintf(tmp_buffer, MAX_STRING_LENGTH,
	     "%08d_ecmwf_%02d.nc", data.vdat[0], data.station_id);
    tmp_buffer[MAX_STRING_LENGTH-1] = '\0';
    handle_error(nc_create(tmp_buffer, NC_CLOBBER, &ncid));
    if (!quiet) {
      fprintf(stderr, "Writing %s\n", tmp_buffer);
    }

    /* Global attributes */
    add_string_attribute(ncid, NC_GLOBAL, "Conventions", "CF-1.0");
    
    add_string_attribute(ncid, NC_GLOBAL, "title",
			 "ECMWF single-site model output");

    if (nfreqs) {
      nct_append_attribute(ncid, NC_GLOBAL, "title",
		   " and corresponding microwave propagation parameters", "");
    }

    if (data.station_name) {
      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "location",
				   strlen(data.station_name),
				   data.station_name));
      nct_append_attribute(ncid, NC_GLOBAL, "title", " over ", "");
      nct_append_attribute(ncid, NC_GLOBAL, "title", data.station_name, "");
    }

    add_string_attribute(ncid, NC_GLOBAL, "source", "ECMWF model");
    add_string_attribute(ncid, NC_GLOBAL, "institution",
			 "European Centre for Medium Range Weather Forecasts");

    year = data.idat/10000;
    month = data.idat/100 - year*100;
    day = data.idat - year*10000 - month*100;

    hour = data.itim;
    minute = (data.itim-hour)*60.0;
    second = (data.itim-hour-((float)minute)/60.0)*3600;

    snprintf(tmp_buffer, MAX_STRING_LENGTH,
	     "%04d-%02d-%02d %02d:%02d:%02d 00:00",
	     year, month, day, hour, minute, second);
    tmp_buffer[MAX_STRING_LENGTH-1] = '\0';
    add_string_attribute(ncid, NC_GLOBAL, "initialization_time",
			 tmp_buffer);

    /* Add a `history' attribute */
    handle_error(nct_add_history(ncid, "NetCDF created from DDH file", user));

    /* Add a `command_line' attribute */
    handle_error(nct_add_command_line(ncid, argc, argv));

    add_string_attribute(ncid, NC_GLOBAL, "software_version", VERSION);

    /* Add time dimension */
    handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &timedimid));

    /* Add level dimension */
    handle_error(nc_def_dim(ncid, "level", data.nlevels, &leveldimid));

    handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
			    &latid));
    add_attributes(ncid, latid, "degrees_north",
		   "Latitude of model gridpoint", "latitude");

    handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
			    &lonid));
    add_attributes(ncid, lonid, "degrees_east",
		   "Longitude of model gridpoint", "longitude");

    /* Add time coordinate variable */ 
    dimids[0] = timedimid;
    handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
			    &timeid));
    year = data.vdat[0]/10000;
    month = data.vdat[0]/100 - year*100;
    day = data.vdat[0] - year*10000 - month*100;
    snprintf(tmp_buffer, MAX_STRING_LENGTH,
	     "hours since %04d-%02d-%02d 00:00:00 00:00",
	     year, month, day);
    tmp_buffer[MAX_STRING_LENGTH-1] = '\0';
    add_attributes(ncid, timeid, tmp_buffer,
		   "Hours UTC", "time");
    add_string_attribute(ncid, timeid, "axis", "T");

    /* Add level coordinate variable */
    dimids[0] = leveldimid;
    handle_error(nc_def_var(ncid, "level", NC_SHORT, 1,
			    dimids, &levelid));
    add_attributes(ncid, levelid, "1", "Model level",
		   "model_level_number");
    add_string_attribute(ncid, levelid, "axis", "Z");
    add_string_attribute(ncid, levelid, "positive", "down");

    dimids[0] = timedimid;
    dimids[1] = leveldimid;

    add_float_array(ncid, &varid[0], "pressure", dimids, "Pa",
		    "Pressure", "pressure");
    add_float_array(ncid, &varid[1], "height", dimids, "m",
		    "Height above ground", "height");
    add_float_array(ncid, &varid[2], "temperature", dimids, "K",
		    "Temperature", "temperature");
    add_float_array(ncid, &varid[3], "q", dimids, "1",
		    "Specific humidity", "specific_humidity");
    add_float_array(ncid, &varid[4], "rh", dimids, "1",
		    "Relative humidity", "relative_humidity");
    add_string_attribute(ncid, varid[4], COMMENT_NAME,
			 "With respect to ice below 0 degrees C.");
    add_float_array(ncid, &varid[5], "ql", dimids, "1",
		    "Gridbox-mean liquid water mixing ratio",
		    "mass_fraction_of_cloud_liquid_water_in_air");
    add_float_array(ncid, &varid[6], "qi", dimids, "1",
		    "Gridbox-mean ice water mixing ratio",
		    "mass_fraction_of_cloud_ice_in_air");
    add_float_array(ncid, &varid[7], "uwind", dimids, "m s-1",
		    "Zonal wind", "eastward_wind");
    add_float_array(ncid, &varid[8], "vwind", dimids, "m s-1",
		    "Meridional wind", "northward_wind");
    add_float_array(ncid, &varid[9], "wwind", dimids, "m s-1",
		    "Vertical wind", "upward_wind");
    add_float_array(ncid, &varid[10], "fraction", dimids, "1",
		    "Cloud fraction", "cloud_area_fraction");

    if (nfreqs) {
      /* Add frequency dimension */
      handle_error(nc_def_dim(ncid, "frequency", nfreqs, &freqdimid));
      
      /* Add frequency coordinate variable */
      dimids[0] = freqdimid;
      handle_error(nc_def_var(ncid, "frequency", NC_FLOAT, 1,
			      dimids, &freqid));
      add_attributes(ncid, freqid, "GHz", "Frequency", NULL);

      dimids[0] = timedimid;
      dimids[1] = leveldimid;
      dimids[2] = freqdimid;

      add_3D_float_array(ncid, &attenid, "gas_atten", dimids, "dB",
			 "Two-way attenuation from the ground"
			 " due to atmospheric gases", NULL);

      add_3D_float_array(ncid, &attenrateid, "specific_gas_atten",
			 dimids, "dB km-1", "Specific one-way attenuation"
			 " due to atmospheric gases", NULL);

      add_3D_float_array(ncid, &satattenrateid,
			 "specific_saturated_gas_atten",
			 dimids, "dB km-1",
			 "Specific one-way attenuation due to atmospheric"
			 " gases for saturated air (saturated with respect"
			 " to ice below 0 degrees C)", NULL);

      add_3D_float_array(ncid, &K2id, "K2", dimids, "dB km-1",
			 "Dielectric parameter (|K|^2) of liquid water",
			 NULL);

      add_3D_float_array(ncid, &liquidattenrateid, "specific_liquid_atten",
			 dimids, "(dB km-1)/(g m-3)", "Specific one-way"
			 " attenuation due to liquid water, per unit liquid"
			 " water content", NULL);

      handle_error(nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME,
					"The gaseous attenuation parameters"
					" were calculated using "
					LIEBE_COMMENT, "\n"));
      handle_error(nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME,
					"The dielectric parameters of liquid"
					" water were calculated using "
					MANABE_LIEBE_HUFFORD_COMMENT, "\n"));
    }

    handle_error(nc_enddef(ncid));

    /* Set the scalar variables */
    tmp_index = 0;
    handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &data.latitude));
    handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &data.longitude));

    /* Set the levels */
    for (j = 0; j < data.nlevels; j++) {
      short level = data.nlevels - j;
      handle_error(nc_put_var1_short(ncid, levelid, &j, &level));
    }

    if (nfreqs) {
      /* Set the frequencies */
      handle_error(nc_put_var_float(ncid, freqid, freq));
    }

    for (i = 0; i < data.nprofiles; i++) {
      size_t start[3], count[3];
      float tmp_float;

      start[0] = i;
      start[1] = 0;
      start[2] = 0;

      count[0] = 1;
      count[1] = data.nlevels;
      count[2] = 1;

      /* Set the time of this profile */
      if (i && data.vdat[i] > data.vdat[i-1]) {
	hour_mod += 24.0;
      }
      tmp_float = data.vtim[i] + hour_mod;
      handle_error(nc_put_var1_float(ncid, timeid, start, &tmp_float));

/* Problem with I being the imaginary number */
#ifdef I
#undef I
#endif

      /* Do all the model variables */
      handle_error(nc_put_vara_float(ncid, varid[0], start, count, data.P[i]));
      handle_error(nc_put_vara_float(ncid, varid[1], start, count, data.H[i]));
      handle_error(nc_put_vara_float(ncid, varid[2], start, count, data.T[i]));
      handle_error(nc_put_vara_float(ncid, varid[3], start, count, data.Q[i]));
      handle_error(nc_put_vara_float(ncid, varid[4], start, count, data.R[i]));
      handle_error(nc_put_vara_float(ncid, varid[5], start, count, data.L[i]));
      handle_error(nc_put_vara_float(ncid, varid[6], start, count, data.I[i]));
      handle_error(nc_put_vara_float(ncid, varid[7], start, count, data.U[i]));
      handle_error(nc_put_vara_float(ncid, varid[8], start, count, data.V[i]));
      handle_error(nc_put_vara_float(ncid, varid[9], start, count, data.W[i]));
      handle_error(nc_put_vara_float(ncid, varid[10],start, count, data.A[i]));

      if (nfreqs) {
	/* Get attenuation profiles */
	for (j = 0; j < nfreqs; j++) {
	  float atten[100];
	  float atten_rate[100];
#define K2 atten
#define liquid_atten atten_rate

	  start[2] = j;

	  ddh_attenuation_profile(&data, freq[j], i, atten, atten_rate);
	  handle_error(nc_put_vara_float(ncid, attenid, start, count,
					 atten));
	  handle_error(nc_put_vara_float(ncid, attenrateid, start, count,
					 atten_rate));

	  ddh_saturated_attenuation_profile(&data, freq[j], i, atten, atten_rate);
	  handle_error(nc_put_vara_float(ncid, satattenrateid, start, count,
					 atten_rate));

	  ddh_liquid_profile(&data, freq[j], i, K2, liquid_atten);
	  handle_error(nc_put_vara_float(ncid, K2id, start, count, K2));
	  handle_error(nc_put_vara_float(ncid, liquidattenrateid,
					 start, count, liquid_atten));
	}
      }
    }
    nc_close(ncid);
  }
  exit(0);
}
