#include <stdio.h>
#include <math.h>

#include "sonde.h"

static T_ERROR ERR_OPEN = {1, "Could not open file", TRUE};
#define DEBUG fprintf(stderr, "Line: %d of %s\n",  __LINE__, __FILE__)

static
void
__interpolate(float *x, float *y, int n, float threshold)
{
  int i = 0, ilast;

  /* Find first valid number */
  while (i < n && y[i] <= threshold) {
    i++;
  }

  if (i >= n) {
    return;
  }
  else if (i) {
    /* Set the first batch of invalid numbers to the first valid number */
    int k;
    for (k = 0; k < i; k++) {
      y[k] = y[i];
    }
  }

  ilast = i;

  while (++i < n) {
    if (y[i] > threshold) {
      if (i-ilast > 1 && x[ilast] - x[i] > 1.0) {
	/* Do the interpolation (provided there is a finite pressure difference) */
	float scaling = 1.0 / (x[ilast] - x[i]);
	int k;
	for (k = ilast+1; k < i; k++) {
	  y[k] = scaling * (y[ilast] * (x[k] - x[i])
			    + y[i] * (x[ilast] - x[k]));
	}
      }
      ilast = i;
    }
    else {
      /* The final points are set to be the same as the last valid points */
       y[i] = y[ilast];
    }
  }
}

static
void
__interpolate_vector(float *x, float *dir, float *y, int n, float threshold)
{
  int i = 0, ilast;

  /* Find first valid number */
  while (i < n && y[i] <= threshold) {
    i++;
  }

  if (i >= n) {
    return;
  }
  else if (i) {
    /* Set the first batch of invalid numbers to the first valid number */
    int k;
    for (k = 0; k < i; k++) {
      y[k] = y[i];
      dir[k] = dir[i];
    }
  }

  ilast = i;

#define PI_OVER_180 0.0174532925199433
#define X180_OVER_PI 57.2957795130823

  while (++i < n) {
    if (y[i] > threshold) {
      if (i-ilast > 1 && x[ilast] - x[i] > 1.0) {
	/* Do the interpolation (provided there is a finite pressure difference) */
	float scaling = 1.0 / (x[ilast] - x[i]);
	float uilast = y[ilast] * sin(dir[ilast] * PI_OVER_180);
	float vilast = y[ilast] * cos(dir[ilast] * PI_OVER_180);
	float ui = y[i] * sin(dir[i] * PI_OVER_180);
	float vi = y[i] * cos(dir[i] * PI_OVER_180);
	int k;
	for (k = ilast+1; k < i; k++) {
	  float u = scaling * (uilast * (x[k] - x[i])
			       + ui * (x[ilast] - x[k]));
	  float v = scaling * (vilast * (x[k] - x[i])
			       + vi * (x[ilast] - x[k]));
	  y[k] = sqrt(u*u + v*v);
	  dir[k] = atan2(u, v) * X180_OVER_PI;
	  if (dir[k] < 0.0) {
	    dir[k] += 360.0;
	  }
	}
      }
      ilast = i;
    }
    else {
      /* The final points are set to be the same as the last valid points */
       y[i] = y[ilast];
       dir[i] = dir[ilast];
    }
  }
}

T_ERROR *
sonde_load(const char *filename, SondeAscent *ascent)
{
  FILE *sondefile;
  char buffer[SONDE_MAX_LINE_LENGTH];
  int format = 1; /* 1 = new, 2 = old */
  int i = 0;

  sondefile = fopen(filename, "r");
  if (!sondefile) {
    return &ERR_OPEN;
  }

  ascent->start = 0;

  switch (format) {
  case 1:
    fgets(buffer, SONDE_MAX_LINE_LENGTH, sondefile);
    sscanf(buffer, "%*6s%4d%2d%2d %d %f %f",
	   &(ascent->year), &(ascent->month), &(ascent->day),
	   &(ascent->z),
	   &(ascent->latitude), &(ascent->longitude));

    fgets(buffer, SONDE_MAX_LINE_LENGTH, sondefile);
    while(fgets(buffer, SONDE_MAX_LINE_LENGTH, sondefile) != NULL
	  && strlen(buffer) > 57) {
      sscanf(buffer, "%9f%9f%10f%10f%9f%10f",
	     &(ascent->pressure[i]),
	     &(ascent->height[i]),
	     &(ascent->temperature[i]),
	     &(ascent->dew_point[i]),
	     &(ascent->wind_direction[i]),
	     &(ascent->wind_speed[i]));
      if (i && ascent->pressure[i] > ascent->pressure[i-1]) {
	ascent->start = i;
      }
      i++;
    }
    ascent->n = i;
    break;

  case 2:
    ascent->latitude = -999.0;
    ascent->longitude = -999.0;
    fgets(buffer, SONDE_MAX_LINE_LENGTH, sondefile);
    while(fgets(buffer, SONDE_MAX_LINE_LENGTH, sondefile) != NULL ) {
      sscanf(buffer, "%6f%6f%6f%6f%*12s%6f%6f",
	     &(ascent->pressure[i]),
	     &(ascent->height[i]),
	     &(ascent->temperature[i]),
	     &(ascent->dew_point[i]),
	     &(ascent->wind_direction[i]),
	     &(ascent->wind_speed[i]));
      ascent->pressure[i] *= 10.0;
      ascent->temperature[i] = ascent->temperature[i] * 0.1 + 273.16;
      ascent->dew_point[i] = ascent->dew_point[i] * 0.1 + 273.16;
      ascent->wind_speed[i] /= 1.944;
      if (i && ascent->pressure[i] > ascent->pressure[i-1]) {
	ascent->start = i;
      }
      i++;
    }
    ascent->n=i;
    break;
  }
  fclose(sondefile);
  return NULL;
}

T_ERROR *
sonde_interpolate(SondeAscent *ascent)
{
  int i, j=0;
  __interpolate(ascent->pressure, ascent->height, ascent->n, SONDE_THRESHOLD_VALUE);
  __interpolate(ascent->pressure, ascent->dew_point, ascent->n, SONDE_THRESHOLD_VALUE);
  __interpolate(ascent->pressure, ascent->temperature, ascent->n, SONDE_THRESHOLD_VALUE);
  __interpolate_vector(ascent->pressure, ascent->wind_direction, ascent->wind_speed,
		       ascent->n, SONDE_THRESHOLD_WIND);

  /* Remove duplicate values */
  for (i = 1; i < ascent->n; i++) {
    if (ascent->height[i] - ascent->height[j] > 1.0) {
      /* Found independent value */
      j++;
      if (j < i) {
	/* Shuffle the values up */
	ascent->height[j] = ascent->height[i];
	ascent->temperature[j] = ascent->temperature[i];
	ascent->dew_point[j] = ascent->dew_point[i];
	ascent->wind_direction[j] = ascent->wind_direction[i];
	ascent->wind_speed[j] = ascent->wind_speed[i];
	ascent->pressure[j] = ascent->pressure[i];
      }
    }
  }
  ascent->n = j + 1;  
  return NULL;
}
